#ifndef IMP_HPP_INCLUDED
#define IMP_HPP_INCLUDED


#include<cstdint>
#include<cstdio>


enum class Dir
{
  top,
  bottom,
  left,
  right,
  front,
  back,

  END,

};


struct
Point3D
{
  int  x;
  int  y;
  int  z;

  constexpr Point3D(int  x_=0, int  y_=0, int  z_=0):
  x(x_), y(y_), z(z_)
  {
  }


  void  reset(int  x_=0, int  y_=0, int  z_=0)
  {
    x = x_;
    y = y_;
    z = z_;
  }

  Point3D  operator+(const Point3D&  rhs) const{return Point3D(x+rhs.x,y+rhs.y,z+rhs.z);}
  Point3D  operator-(const Point3D&  rhs) const{return Point3D(x-rhs.x,y-rhs.y,z-rhs.z);}

  Point3D&  operator+=(const Point3D&  rhs)
  {
    x += rhs.x;
    y += rhs.y;
    z += rhs.z;

    return *this;
  }

  Point3D&  operator-=(const Point3D&  rhs)
  {
    x -= rhs.x;
    y -= rhs.y;
    z -= rhs.z;

    return *this;
  }

  bool  operator==(const Point3D&  rhs) const
  {
    return((x == rhs.x) &&
           (y == rhs.y) &&
           (z == rhs.z));
  }

  void  print() const
  {
    printf("[print Point3D] X:%4d, Y:%4d, Z:%4d\n",x,y,z);
  }

};


struct
Point2D
{
  int  x;
  int  y;

  constexpr Point2D(int  x_=0, int  y_=0):
  x(x_), y(y_)
  {
  }


  void  print() const
  {
    printf("[print Point2D] X:%4d, Y:%4d\n",x,y);
  }
};


constexpr Point2D  to_isometric(const Point3D&  pt){return Point2D(pt.x-pt.y,(pt.y/2)+(pt.x/2)+pt.z);}
constexpr Point2D  to_isometric(int  x, int  y, int  z){return Point2D(x-y,(y/2)+(x/2)+z);}

constexpr Point2D  to_topview(  const Point3D&  pt){return Point2D(pt.x,pt.y);}
constexpr Point2D  to_frontview(const Point3D&  pt){return Point2D(pt.x,pt.z);}
constexpr Point2D  to_leftview( const Point3D&  pt){return Point2D(pt.y,pt.z);}


struct
RenderingParameter
{
  int  width ;
  int  height;

  int  x_offset;

  constexpr RenderingParameter(int  x_width, int  y_width, int  z):
     width(to_isometric(x_width,      0,0).x*2),
    height(to_isometric(x_width,y_width,z).y  ),
  x_offset(to_isometric(x_width,      0,0).x  )
  {
  }


  constexpr int  get_areasize() const
  {
    return(width*height);
  }

};




#endif




