#include"node.hpp"
#include"field.hpp"




Node::
Node(const std::u16string&  name):
name(name),
parent(nullptr),
x_width(8),
y_width(8),
z_width(8)
{
  source.offset.reset(20,20,20);

  remake();
}


Node::
~Node()
{
  delete_children();
}




void
Node::
delete_children()
{
    for(auto  child: children)
    {
      delete child;
    }


  children.clear();
}


void
Node::
remove_children()
{
  children.clear();
}


void
Node::
append(Node&  child)
{
  child.parent = this;

  children.emplace_back(&child);
}


void
Node::
remake()
{
  polygons.clear();

  polygons.emplace_back(Point3D(0,0,0),Point3D(x_width,0,0),Point3D(0,y_width,0));
  polygons.emplace_back(Point3D(x_width,0,0),Point3D(0,y_width,0),Point3D(x_width,y_width,0));

  polygons.emplace_back(Point3D(0,0,z_width),Point3D(x_width,0,z_width),Point3D(0,y_width,z_width));
  polygons.emplace_back(Point3D(x_width,0,z_width),Point3D(0,y_width,z_width),Point3D(x_width,y_width,z_width));


  polygons.emplace_back(Point3D(0,y_width,0),Point3D(x_width,y_width,0),Point3D(0,y_width,z_width));
  polygons.emplace_back(Point3D(x_width,y_width,0),Point3D(0,y_width,z_width),Point3D(x_width,y_width,z_width));

  polygons.emplace_back(Point3D(0,0,0),Point3D(x_width,0,0),Point3D(0,0,z_width));
  polygons.emplace_back(Point3D(x_width,0,0),Point3D(0,0,z_width),Point3D(x_width,0,z_width));


  polygons.emplace_back(Point3D(x_width,y_width,0),Point3D(x_width,0,0),Point3D(x_width,y_width,z_width));
  polygons.emplace_back(Point3D(x_width,0,0),Point3D(x_width,y_width,z_width),Point3D(x_width,0,z_width));

  polygons.emplace_back(Point3D(0,y_width,0),Point3D(0,0,0),Point3D(0,y_width,z_width));
  polygons.emplace_back(Point3D(0,0,0),Point3D(0,y_width,z_width),Point3D(0,0,z_width));
}


void
Node::
write_as_root()
{
  applied.offset = source.offset;
  applied.center = source.offset+source.center;
  applied.x_degree = source.x_degree;
  applied.y_degree = source.y_degree;
  applied.z_degree = source.z_degree;

    for(auto&  src_poly: polygons)
    {
      Polygon3D  tmp = src_poly;

      tmp += applied.offset;

      rotate(tmp.a_point,applied);
      rotate(tmp.b_point,applied);
      rotate(tmp.c_point,applied);

      FieldRenderer::push(tmp);
    }


    for(auto  child: children)
    {
      child->write();
    }
}


void
Node::
write()
{
  applied.offset = parent->applied.center+source.offset;
  applied.center = applied.offset+source.center;

  rotate(applied.offset,parent->applied);
  rotate(applied.center,parent->applied);


  applied.x_degree = -parent->applied.x_degree;
  applied.y_degree = -parent->applied.y_degree;
  applied.z_degree = -parent->applied.z_degree;

  rotate(applied.offset,applied);

  applied.x_degree = parent->applied.x_degree+source.x_degree;
  applied.y_degree = parent->applied.y_degree+source.y_degree;
  applied.z_degree = parent->applied.z_degree+source.z_degree;

    for(auto&  src_poly: polygons)
    {
      Polygon3D  tmp = src_poly;

      tmp += applied.offset;

      rotate(tmp.a_point,applied);
      rotate(tmp.b_point,applied);
      rotate(tmp.c_point,applied);

      FieldRenderer::push(tmp);
    }


    for(auto&  child: children)
    {
      child->write();
    }
}


void
Node::
render_center(Drawer&  dr) const
{
  constexpr int  r = 32;

  const int  x = applied.center.x;
  const int  y = applied.center.y;
  const int  z = applied.center.z;

  ObjectRenderer::draw_line(dr,0xFFFFFFFF,Point3D(x-r,y  ,z  ),Point3D(x+r,y  ,z  ));
  ObjectRenderer::draw_line(dr,0xFFFFFFFF,Point3D(x  ,y-r,z  ),Point3D(x  ,y+r,z  ));
  ObjectRenderer::draw_line(dr,0xFFFFFFFF,Point3D(x  ,y  ,z-r),Point3D(x  ,y  ,z+r));
}


void
Node::
render_lines(Drawer&  dr, uint8_t  l) const
{
  auto  pixel = to_pixel(l,l,l);

    if(parent)
    {
      ObjectRenderer::draw_circle(dr,pixel,parent->applied.center,8);

      ObjectRenderer::draw_line(dr,pixel,parent->applied.center,applied.center);
    }


  ObjectRenderer::draw_circle(dr,pixel,applied.center,12);

    for(auto&  child: children)
    {
      child->render_lines(dr,l-48);
    }
}




void
Node::
write(Stream&  s, const Parameter&  param)
{
  Polygon3D::write(s,param.offset);
  Polygon3D::write(s,param.center);

  s.put_vlint(param.x_degree);
  s.put_vlint(param.y_degree);
  s.put_vlint(param.z_degree);
}


void
Node::
read(Stream&  s, Parameter&  param)
{
  Polygon3D::read(s,param.offset);
  Polygon3D::read(s,param.center);

  param.x_degree = s.get_vlint();
  param.y_degree = s.get_vlint();
  param.z_degree = s.get_vlint();
}




void
Node::
write(Stream&  s) const
{
  s.put_vlint(number);

  s.write(name);

  s.put_vlint(x_width);
  s.put_vlint(y_width);
  s.put_vlint(z_width);

  s.put_uvlint(polygons.size());

    for(auto&  poly: polygons)
    {
      poly.write(s);
    }


  write(s,source);
  write(s,applied);

  s.put_uvlint(children.size());

    for(auto  child: children)
    {
      child->write(s);
    }
}


void
Node::
read(Stream&  s)
{
  number = s.get_vlint();

  name.clear();

  s.read(name);

  x_width = s.get_vlint();
  y_width = s.get_vlint();
  z_width = s.get_vlint();

  int  n = s.get_uvlint();

  polygons.resize(n);

    for(auto&  poly: polygons)
    {
      poly.read(s);
    }


  remake();

  read(s,source);
  read(s,applied);

  n = s.get_uvlint();

  delete_children();

    while(n--)
    {
      auto  child = new Node;

      child->read(s);

      append(*child);
    }
}


void
Node::
print() const
{
  printf("[%s] width(%3d,%3d,%3d),offset(%3d,%3d,%3d)\n",
to_string(name).data(),
x_width,
y_width,
z_width,
source.offset.x,
source.offset.y,
source.offset.z);

    for(auto  child: children)
    {
//      child->print();
    }
}





