#include"object.hpp"
#include"field.hpp"




namespace{


void
draw_point(Drawer&  dr, pixel32_t  pixel, int  x, int  y)
{
    if((x >=               0) &&
       (x <   dr.get_width()) &&
       (y >=               0) &&
       (y <  dr.get_height()))
    {
      dr.get_pixel(x,y) = pixel;
    }
}


void
draw_point(Drawer&  dr, pixel32_t  pixel, int  x, int  y, int  z)
{
  Point2D  pt = to_isometric(x,y,z);

  pt.x += Field::get_rendparam().x_offset;

  draw_point(dr,pixel,pt.x,pt.y);
}


}




void
ObjectRenderer::
draw_line(Drawer&  dr, pixel32_t  pixel, int x0, int y0, int x1, int y1)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
   int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
   int err = dx+dy, e2; /* error value e_xy */
 
   for(;;){  /* loop */
      draw_point(dr,pixel,x0,y0);
      if (x0==x1 && y0==y1) break;
      e2 = 2*err;
      if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
      if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
   }
}


void
ObjectRenderer::
draw_line(Drawer&  dr, pixel32_t  pixel, int  x1, int  y1, int  z1, int  x2, int  y2, int  z2)
{
  Point2D  a = to_isometric(x1,y1,z1);
  Point2D  b = to_isometric(x2,y2,z2);

  constexpr int  x_base = Field::get_rendparam().x_offset;

  draw_line(dr,pixel,x_base+a.x,
                            a.y,
                     x_base+b.x,
                            b.y);
}


void
ObjectRenderer::
draw_line(Drawer&  dr, pixel32_t  pixel, const Point3D&  a, const Point3D&  b)
{
  draw_line(dr,pixel,a.x,
                     a.y,
                     a.z,
                     b.x,
                     b.y,
                     b.z);
}


void
ObjectRenderer::
draw_circle(Drawer&  dr, pixel32_t  pixel, int xm, int ym, int r)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int x = -r, y = 0, err = 2-2*r; /* II. Quadrant */ 
   do {
      draw_point(dr,pixel,xm-x, ym+y); /*   I. Quadrant */
      draw_point(dr,pixel,xm-y, ym-x); /*  II. Quadrant */
      draw_point(dr,pixel,xm+x, ym-y); /* III. Quadrant */
      draw_point(dr,pixel,xm+y, ym+x); /*  IV. Quadrant */
      r = err;
      if (r <= y) err += ++y*2+1;           /* e_xy+e_y < 0 */
      if (r > x || err > y) err += ++x*2+1; /* e_xy+e_x > 0 or no 2nd y-step */
   } while (x < 0);
}


void
ObjectRenderer::
draw_circle(Drawer&  dr, pixel32_t  pixel, const Point3D&  point, int r)
{
  Point2D  pt2d = to_isometric(point);

  constexpr int  x_base = Field::get_rendparam().x_offset;

  draw_circle(dr,pixel,x_base+pt2d.x,
                              pt2d.y,r);
}




void
ObjectRenderer::
draw_square(Drawer&  dr, pixel32_t  pixel, const Point3D&  point, int r)
{
  Point2D  pt2d = to_isometric(point);

  constexpr int  x_base = Field::get_rendparam().x_offset;

  dr.draw_rect(pixel,x_base+pt2d.x-r,
                            pt2d.y-r,r*2,r*2);
}


void
ObjectRenderer::
draw_ellipse_rect(Drawer&  dr, pixel32_t  pixel, int x0, int y0, int x1, int y1)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int a = abs(x1-x0), b = abs(y1-y0), b1 = b&1; /* values of diameter */
   long dx = 4*(1-a)*b*b, dy = 4*(b1+1)*a*a; /* error increment */
   long err = dx+dy+b1*a*a, e2; /* error of 1.step */

   if (x0 > x1) { x0 = x1; x1 += a; } /* if called with swapped points */
   if (y0 > y1) y0 = y1; /* .. exchange them */
   y0 += (b+1)/2; y1 = y0-b1;   /* starting pixel */
   a *= 8*a; b1 = 8*b*b;

   do {
       draw_point(dr,pixel,x1, y0); /*   I. Quadrant */
       draw_point(dr,pixel,x0, y0); /*  II. Quadrant */
       draw_point(dr,pixel,x0, y1); /* III. Quadrant */
       draw_point(dr,pixel,x1, y1); /*  IV. Quadrant */
       e2 = 2*err;
       if (e2 <= dy) { y0++; y1--; err += dy += a; }  /* y step */ 
       if (e2 >= dx || 2*err > dy) { x0++; x1--; err += dx += b1; } /* x step */
   } while (x0 <= x1);
   
   while (y0-y1 < b) {  /* too early stop of flat ellipses a=1 */
       draw_point(dr,pixel,x0-1, y0); /* -> finish tip of ellipse */
       draw_point(dr,pixel,x1+1, y0++); 
       draw_point(dr,pixel,x0-1, y1);
       draw_point(dr,pixel,x1+1, y1--); 
   }
}




