#ifndef OBJECT_HPP_INCLUDED
#define OBJECT_HPP_INCLUDED


#include"imp.hpp"
#include"drawer.hpp"
#include"controller.hpp"
#include"rep.hpp"
#include"sincos.hpp"
#include"stream.hpp"
#include<list>
#include<algorithm>


template<typename  T>
bool
remove(std::list<T*>&  ls, const T*  const ptr)
{
  auto   it = ls.begin();
  auto  end = ls.end();

    while(it != end)
    {
        if(*it == ptr)
        {
          ls.erase(it);

          return true;
        }


      ++it;
    }


  return false;
}



struct
ObjectRenderer
{
  static void  draw_line(Drawer&  dr, pixel32_t  pixel, int  x1, int  y1, int  x2, int  y2);

  static void  draw_line(Drawer&  dr, pixel32_t  pixel, int  x1, int  y1, int  z1, int  x2, int  y2, int  z2);
  static void  draw_line(Drawer&  dr, pixel32_t  pixel, const Point3D&  a, const Point3D&  b);

  static void  draw_circle(Drawer&  dr, pixel32_t  pixel, int xm, int ym, int r);
  static void  draw_circle(Drawer&  dr, pixel32_t  pixel, const Point3D&  point, int r);
  static void  draw_square(Drawer&  dr, pixel32_t  pixel, const Point3D&  point, int r);

  static void  draw_ellipse_rect(Drawer&  dr, pixel32_t  pixel, int x0, int y0, int x1, int y1);

};


struct
Polygon3D
{
  struct Area{
    int  x_min;
    int  y_min;
    int  z_min;
    int  x_max;
    int  y_max;
    int  z_max;

    static constexpr int  min(int  a, int  b        ){return (a < b)? a:b;}
    static constexpr int  min(int  a, int  b, int  c){return min(a,min(b,c));}
    static constexpr int  max(int  a, int  b        ){return (a > b)? a:b;}
    static constexpr int  max(int  a, int  b, int  c){return max(a,max(b,c));}

    Area(const Polygon3D&  src):
    x_min(min(src.a_point.x,src.b_point.x,src.c_point.x)),
    x_max(max(src.a_point.x,src.b_point.x,src.c_point.x)),
    y_min(min(src.a_point.y,src.b_point.y,src.c_point.y)),
    y_max(max(src.a_point.y,src.b_point.y,src.c_point.y)),
    z_min(min(src.a_point.z,src.b_point.z,src.c_point.z)),
    z_max(max(src.a_point.z,src.b_point.z,src.c_point.z)){}

  };


  struct RenderingData{
    Point2D     top_point;
    Point2D  middle_point;
    Point2D  bottom_point;

    uint32_t  priority_number;

    uint32_t  color;

    RenderingData(){}
    RenderingData(const Polygon3D&  src);

    void     assign(const Polygon3D&  src);
    void  operator=(const Polygon3D&  src);

    void  reset_three_points(const Point2D&  top, const Point2D&  middle, const Point2D&  bottom);

    void  draw(Drawer&  dr) const;

  };


  Point3D  a_point;
  Point3D  b_point;
  Point3D  c_point;

  pixel16_t  pixel;

  Polygon3D(){}

  Polygon3D(const Point3D&  a, const Point3D&  b, const Point3D&  c, pixel16_t  pixel_=to_pixel16(15,15,15)):
  a_point(a),
  b_point(b),
  c_point(c),
  pixel(pixel_){}

  Polygon3D&  operator+=(const Point3D&  offset);

  static void  write(Stream&  s, const Point3D&  pt);
  static void   read(Stream&  s,       Point3D&  pt);

  void  write(Stream&  s) const;
  void   read(Stream&  s)      ;

  void  print() const;

};




struct AssemblingField;
struct    LinkingField;


struct Side;


struct
Vertex: public Point3D
{
  std::list<Side*>  side_list;

};


struct
Side
{
  Vertex&  a;
  Vertex&  b;

  Side(Vertex&  a_, Vertex&  b_): a(a_), b(b_)
  {
    a.side_list.emplace_back(this);
    b.side_list.emplace_back(this);
  }

};




#endif




