#include"object.hpp"
#include"field.hpp"
#include<limits>
#include<cmath>




namespace{


struct
Memo
{
  uint16_t  tmp;
  uint16_t  min;
  uint16_t  max;

  void  reset_min(uint16_t  v){min = std::min(min,v);}
  void  reset_max(uint16_t  v){max = std::max(max,v);}

};


Memo
memobuf[Field::get_rendparam().height];


void
clear_memo(int  y, int  bottom)
{
    if(y < 0)
    {
      y = 0;
    }


    while(y <= bottom)
    {
        if(y >= Field::get_rendparam().height)
        {
          break;
        }


      auto&  m = memobuf[y++];

      m.min = std::numeric_limits<uint16_t>::max();
      m.max = std::numeric_limits<uint16_t>::min();
    }
}


void
set_point_tmp(int  x, int  y)
{
    if((x >= 0) &&
       (x <  Field::get_rendparam().width) &&
       (y >= 0) &&
       (y <  Field::get_rendparam().height))
    {
      memobuf[y].tmp = x;
    }
}


void
set_point(int  x, int  y)
{
    if((x >= 0) &&
       (x <  Field::get_rendparam().width) &&
       (y >= 0) &&
       (y <  Field::get_rendparam().height))
    {
      auto&  m = memobuf[y];

        if(m.tmp <= x){m.reset_max(x);  m.reset_min(m.tmp);}
      else            {m.reset_min(x);  m.reset_max(m.tmp);}
    }
}




void
draw_line(void  (*set)(int,int), int x0, int y0, int x1, int y1)
{
//http://members.chello.at/~easyfilter/bresenham.html
   int dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
   int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
   int err = dx+dy, e2; /* error value e_xy */
 
   for(;;){  /* loop */
      set(x0,y0);
      if (x0==x1 && y0==y1) break;
      e2 = 2*err;
      if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
      if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
   }
}


}




int
get_l(double  a, double  b)
{
  constexpr double  pi = 3.141592653589793;

  auto  v = static_cast<int>(std::atan(a/b)*180/pi);

  return 64-(((64<<10)/90*v)>>10);
}



Polygon3D::RenderingData::
RenderingData(const Polygon3D&  src)
{
  assign(src);
}




void
Polygon3D::RenderingData::
reset_three_points(const Point2D&  top, const Point2D&  middle, const Point2D&  bottom)
{
  top_point    =    top;
  middle_point = middle;
  bottom_point = bottom;
}


void
Polygon3D::RenderingData::
operator=(const Polygon3D&  src)
{
  assign(src);
}


void
Polygon3D::RenderingData::
assign(const Polygon3D&  src)
{
  unsigned int  l = 63;

  unsigned int  r_ = /*to_r4(src.pixel)*/14<<4;
  unsigned int  g_ = /*to_g4(src.pixel)*/12<<4;
  unsigned int  b_ = /*to_b4(src.pixel)*/6<<4;


  const Area  area(src);

  const int  x_width = (area.x_max-area.x_min);
  const int  y_width = (area.y_max-area.y_min);
  const int  z_width = (area.z_max-area.z_min);

  priority_number = (area.x_min+(x_width/2))
                   +(area.y_min+(y_width/2))
                   -(area.z_min+(z_width/2));

  l += get_l(z_width,x_width);
  l += get_l(z_width,y_width);
  l += get_l(y_width,x_width);


  color = to_pixel((r_*l)>>8,(g_*l)>>8,(b_*l)>>8);

  Point2D  a = to_isometric(src.a_point);
  Point2D  b = to_isometric(src.b_point);
  Point2D  c = to_isometric(src.c_point);

    if(a.y <= b.y)
    {
        if(a.y <= c.y)
        {
            if(b.y <= c.y){reset_three_points(a,b,c);}
          else            {reset_three_points(a,c,b);}
        }

      else
        {
          reset_three_points(c,a,b);
        }
    }

  else
    {
        if(b.y <= c.y)
        {
            if(a.y <= c.y){reset_three_points(b,a,c);}
          else            {reset_three_points(b,c,a);}
        }

      else
        {
          reset_three_points(c,b,a);
        }
    }
}


void
Polygon3D::RenderingData::
draw(Drawer&  dr) const
{
  constexpr int  x_base = Field::get_rendparam().x_offset;

  clear_memo(top_point.y,bottom_point.y);

  draw_line(set_point_tmp,x_base+top_point.x   ,top_point.y   ,x_base+bottom_point.x,bottom_point.y);
  draw_line(set_point    ,x_base+top_point.x   ,top_point.y   ,x_base+middle_point.x,middle_point.y);
  draw_line(set_point    ,x_base+middle_point.x,middle_point.y,x_base+bottom_point.x,bottom_point.y);

    for(int  y = top_point.y;  y <= bottom_point.y;  ++y)
    {
        for(int  x = memobuf[y].min;  x <= memobuf[y].max;  ++x)
        {
            if((x >= 0) &&
               (x <  Field::get_rendparam().width) &&
               (y >= 0) &&
               (y <  Field::get_rendparam().height))
            { 
              dr.get_pixel(x,y) = color;
            }
        }
    }
}




Polygon3D&
Polygon3D::
operator+=(const Point3D&  offset)
{
  a_point += offset;
  b_point += offset;
  c_point += offset;

  return *this;
}


void
Polygon3D::
write(Stream&  s, const Point3D&  pt)
{
  s.put_vlint(pt.x);
  s.put_vlint(pt.y);
  s.put_vlint(pt.z);
}


void
Polygon3D::
read(Stream&  s, Point3D&  pt)
{
  pt.x = s.get_vlint();
  pt.y = s.get_vlint();
  pt.z = s.get_vlint();
}


void
Polygon3D::
write(Stream&  s) const
{
  write(s,a_point);
  write(s,b_point);
  write(s,c_point);

  s.put_vlint(pixel);
}


void
Polygon3D::
read(Stream&  s)
{
  read(s,a_point);
  read(s,b_point);
  read(s,c_point);

  pixel = s.get_vlint();
}


void
Polygon3D::
print() const
{
  a_point.print();
  b_point.print();
  c_point.print();
}




